/*
   This file contains a number of pseudo-code implementations for
   hardware based beginCriticalSection() and endCriticalSection()
   functions.

   These implementations were taken from a number of different
   operating systems textbooks.
*/



/*
    page 279 from "An Operating Systems Vade Mecum", 2nd Ed., by Raphael Finkel
*/
void beginCriticalSection()
{
   disableInterrupts();
}


void endCriticalSection()
{
   enableInterrupts();
}



/*
   page 300 from "Operating Systems", 3rd Ed, by Gary Nutt
   page 224 from "Operating Systems: A Modern Perspective", 2nd Ed, by Gary Nutt
*/
void beginCriticalSection(boolean* lock)
{
   disableInterrupts();
   while(*lock)
   {
      // allow an interrupt to occur
      enableInterrupts();
      disableInterrupts();
   }
   *lock = true;
   enableInterrupts();
}


void endCriticalSection(boolean* lock)
{
   *lock = false;
}



/*
   page 214 from "Operating Systems: Internals and Design Principles", 5th Ed,
   by William Stallings
*/
int lock = 0;

void beginCriticalSection()
{
   while( !testAndSet(lock) ) /* empty */ ;  // busy loop

   //while( !testAndSet(lock) ) Sleep(0) ;   // not quite busy loop
}


void endCriticalSection()
{
   lock = 0;
}


// This function should be an atomic machine instruction
// in the CPU's instruction set.
// For example, you could use the Intel IA-32 instruction BTS.
boolean testAndSet(int lock)
{
   if (lock == 0)
   {
      lock = 1;
      return true;
   }
   else
   {
      return false;
   }
}



/*
   page 218 from "Operating Systems: Internals and Design Principles", 6th Ed,
   by William Stallings
*/
int lock = 0;

void beginCriticalSection()
{
   register int key = 1;
   do
      exchange(key, &lock)
   while ( key != 0 );
}


void endCriticalSection()
{
   lock = 0;
}


// This function should be an atomic machine instruction
// in the CPU's instruction set.
// For example, you could use the Intel IA-32 instruction XCHG.
void exchange(int register, int* memory)
{
   int temp = *memory;
   *memory = register;
   register = temp;
}



/*
   page 218 from "Operating Systems: Internals and Design Principles", 6th Ed,
   by William Stallings
*/
int lock = 0;

void beginCriticalSection()
{
   while( 1 == compareAndSwap(lock, 0, 1) ) /* empty */ ;  // busy loop
}


void endCriticalSection()
{
   lock = 0;
}


// This function should be an atomic machine instruction
// in the CPU's instruction set.
int compareAndSwap(int* lock_p, int testValue, int newValue)
{
   int oldValue;
   oldValue = *lock_p;
   if (oldValue == testValue)
   {
      *lock_p = newValue;
   }
   return oldValue
}
