/*
   This file shows how to implement a mutex like object using the Pentium
   machine instruction XCHG. The XCHG instruction exchanges the value stored
   in a memory location with the value stored in a register. This operation,
   which includes reading from the memory location and then writing to the
   memory location, is "atomic", that is, the read and write steps cannot
   be interrupted by an interrupt. So there cannot be a race condition
   caused by the XCHG instruction itself.

   This version of the mutex tries to be more efficient than the last
   version. It replaces the busy loop with a more efficient loop that
   gives up the processor as soon as it detects that it is blocked.

   This implementation works correctly on multicore and multi-processor
   systems since the XCHG instruction implements the processors bus
   locking protocol.
*/
#include <stdlib.h>

typedef int* MyMutex;

MyMutex CreateMyMutex(int initialValue)
{
   int* temp_p;
   temp_p = (int*)malloc(sizeof(int));
   *temp_p = initialValue;
   return temp_p;
}//CreateMyMutex()


void wait(MyMutex mutex_p)
{
   int key = 1;
// loop (efficiently) until the mutex is read as a 0
again: __asm mov  eax, key     // move the key into register eax
       __asm mov  ebx, mutex_p // move the lock pointer into register ebx
       __asm xchg eax, [ebx]   // exchange the key and lock values
       __asm and  eax, eax     // test the previous value of the lock
       __asm jz   done         // exit, since the lock was zero
       Sleep(0);               // we are blocked, so yield the processor
       __asm jmp again         // keep trying till the lock was zero
done:  __asm nop
       return;
}//wait()


void release(MyMutex mutex_p)
{
   *mutex_p = 0;  // unlock the mutex by setting it to 0
   return;
}//release()